"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const graphql_1 = require("graphql");
const ts = require("typescript");
const date_1 = require("./date");
const ts_type_ast_1 = require("ts-type-ast");
function createSchema(fileName, options = {}) {
    const customScalarsMap = new Map([date_1.DateType, ...(options.customScalars || [])].map(it => [it.name, it]));
    const program = ts.createProgram({ options: { strict: true }, rootNames: [fileName] });
    const checker = program.getTypeChecker();
    const sourceFile = program.getSourceFile(fileName);
    const types = ts_type_ast_1.typeAST(checker, sourceFile);
    const map = new Map();
    let anonTypeIdx = 0;
    const schema = createSchemaFromTypes();
    // console.log(printSchema(schema));
    return schema;
    function createSchemaFromTypes() {
        let query;
        let mutation;
        for (let i = 0; i < types.length; i++) {
            const gqlType = createGQL(types[i]);
            if (gqlType instanceof graphql_1.GraphQLObjectType) {
                if (gqlType.name.match(/Query$/)) {
                    query = gqlType;
                }
                if (gqlType.name.match(/Mutation$/)) {
                    mutation = gqlType;
                }
            }
        }
        return new graphql_1.GraphQLSchema({
            query: query,
            mutation: mutation,
        });
    }
    function add(type, gqltype) {
        map.set(type, gqltype);
        return gqltype;
    }
    function createGQL(type) {
        const gqlType = map.get(type);
        if (gqlType)
            return gqlType;
        switch (type.kind) {
            case 'interface':
            case 'interfaceLiteral':
                return createGQLType(type);
            // case 'enum':
            // return add(type, createGQLEnum(type));
            case 'union':
                return add(type, createGQLUnion(type));
            case 'array':
                return new graphql_1.GraphQLList(add(type, nullable(false, createGQL(type.element))));
            case 'native':
                if (type.name === 'Date') {
                    return nonNull(customScalarsMap.get('Date'));
                }
                throw new Error('Unexpected type: ' + type.name);
            case 'primitive':
                return add(type, createGQLPrimitive(type));
        }
        throw new Error('Unexpected type: ' + JSON.stringify(type));
    }
    function nullable(nullable, type) {
        return nullable ? type : new graphql_1.GraphQLNonNull(type);
    }
    function createGQLType(type) {
        let typeName = type.kind === 'interface' ? type.name : '';
        const isInput = typeName && typeName.match(/Input$/);
        const Class = isInput ? graphql_1.GraphQLInputObjectType : graphql_1.GraphQLObjectType;
        const fields = {};
        if (type.kind === 'interfaceLiteral') {
            for (let i = 0; i < type.members.length; i++) {
                const member = type.members[i];
                // console.log(member.name, member.type);
                if (member.name === '__typename' &&
                    member.type.kind === 'primitive' &&
                    typeof member.type.literal === 'string') {
                    typeName = member.type.literal;
                }
            }
        }
        if (typeName === '')
            typeName = 'Anonymous' + ++anonTypeIdx;
        const gqlType = new Class({
            name: typeName,
            description: type.kind === 'interface' ? type.doc : undefined,
            fields: fields,
        });
        add(type, gqlType);
        type.members.reduce((obj, member) => {
            const memberType = {
                type: nullable(member.optional, createGQL(member.type)),
                args: member.args
                    ? member.args[0].type.members.reduce((acc, arg) => {
                        acc[arg.name] = {
                            description: arg.doc,
                            defaultValue: undefined,
                            type: nullable(arg.optional, createGQL(arg.type)),
                        };
                        return acc;
                    }, {})
                    : undefined,
                // todo:
                deprecationReason: undefined,
                description: member.doc,
            };
            if (member.name !== '__typename') {
                obj[member.name] = memberType;
            }
            return obj;
        }, fields);
        return gqlType;
    }
    function createGQLUnion(type) {
        return new graphql_1.GraphQLUnionType({
            name: type.name,
            description: type.doc,
            types: type.members.map(member => createGQL(member)),
        });
    }
    function createGQLPrimitive(type) {
        if (type.rawType === 'ID')
            return graphql_1.GraphQLID;
        switch (type.type) {
            case 'number':
                return type.rawType === 'Int' ? graphql_1.GraphQLInt : graphql_1.GraphQLFloat;
            case 'string':
                return graphql_1.GraphQLString;
            case 'boolean':
                return graphql_1.GraphQLBoolean;
        }
        const customType = customScalarsMap.get(type.type);
        if (customType)
            return customType;
        throw new Error('Unexpected type: ' + JSON.stringify(type));
    }
}
exports.createSchema = createSchema;
function never(never) {
    throw new Error('Never possible');
}
function nonNull(val) {
    if (val === undefined)
        throw new Error('Undefined is not expected here');
    return val;
}
